#if CONFIG_FREERTOS_UNICORE
#define ARDUINO_RUNNING_CORE 0
#else
#define ARDUINO_RUNNING_CORE 1
#endif

#define MINIMUM_DELAY_ALLOWING_MULTITASKING 50

#include "driver/twai.h"
#include <MD_Parola.h>
#include <MD_MAX72xx.h>

#define DVERT_VERSION    0x01
#define DVERT_SUBVERSION 0x00 /* 20240111 (v1.0)*/

#define CAN_CARRIER_ID 0x00AA
#define CAN_BROADCAST_STOREY_BUTTON_BOARD 0x40
#define CAN_PREFIX_LED_BUTTON_COMMAND 0x100

#define HARDWARE_TYPE MD_MAX72XX::ICSTATION_HW
#define MAX_DEVICES 2 // 2 led matrices
#define CS_PIN 19 // in other previous boards it was 21

#define USW1  13
#define USW2  14
#define USW3  4
#define USW4  16
#define USW5  17
#define USW6  5

#define LED_BRD 32
#define EXT_LED_BT1   2   /* hand wired at first prototype */
#define EXT_LED_BT2   25  /* hand wired at first prototype */

#define BT1 34
#define BT2 35

#define RX_PIN 26
#define TX_PIN 27

#define SDA 21
#define SCL 22

const int nUsw[6] = {USW1, USW2, USW3, USW4, USW5, USW6};
volatile int nBrdAddress;

// create an instance of the MD_Parola class
MD_Parola ledMatrix = MD_Parola(HARDWARE_TYPE, CS_PIN, MAX_DEVICES);

void vSetupIO() {
  int i;

  for (i = 0; i < 6; i++) {
    pinMode(nUsw[i], INPUT_PULLUP);
  }
  pinMode(BT1, INPUT);
  pinMode(BT2, INPUT);
  pinMode(LED_BRD, OUTPUT);
  pinMode(EXT_LED_BT1, OUTPUT);
  pinMode(EXT_LED_BT2, OUTPUT);
  digitalWrite(LED_BRD, HIGH);
}

void vSetupLedMatrix() {
  ledMatrix.begin();         // initialize the object
  ledMatrix.setIntensity(5); // set the brightness of the LED matrix display (from 0 to 15)
  ledMatrix.displayClear();  // clear led matrix display
}

byte byReadAddress() {
  byte byAddr = (digitalRead(USW1)) ? 0x00 : 0x1;

  byAddr |= (digitalRead(USW2)) ? 0x00 : 0x2;
  byAddr |= (digitalRead(USW3)) ? 0x00 : 0x4;
  byAddr |= (digitalRead(USW4)) ? 0x00 : 0x8;
  byAddr |= (digitalRead(USW5)) ? 0x00 : 0x10;
  byAddr |= (digitalRead(USW6)) ? 0x00 : 0x20;

  return byAddr;
}

void vSetupCan() {
  // Serial.setTxTimeoutMs(0);   // prevent slow timeouts
  // pinMode(CAN_RS, OUTPUT);    // INPUT (high impedance) = slope control mode, OUTPUT = see next line
  // digitalWrite(CAN_RS, LOW);  // LOW = high speed mode, HIGH = low power mode (listen only)
  // twai_general_config_t g_config = TWAI_GENERAL_CONFIG_DEFAULT((gpio_num_t)TX_PIN, (gpio_num_t)RX_PIN, TWAI_MODE_LISTEN_ONLY);  // TWAI_MODE_NORMAL, TWAI_MODE_NO_ACK or TWAI_MODE_LISTEN_ONLY
  twai_general_config_t g_config = TWAI_GENERAL_CONFIG_DEFAULT((gpio_num_t)TX_PIN, (gpio_num_t)RX_PIN, TWAI_MODE_NORMAL);  // TWAI_MODE_NORMAL, TWAI_MODE_NO_ACK or TWAI_MODE_LISTEN_ONLY
  // twai_timing_config_t t_config  = TWAI_TIMING_CONFIG_500KBITS();
  twai_timing_config_t t_config  = TWAI_TIMING_CONFIG_100KBITS();
  twai_filter_config_t f_config  = TWAI_FILTER_CONFIG_ACCEPT_ALL();
  twai_driver_install(&g_config, &t_config, &f_config);
  twai_start();
}

void vReadCan() {
  twai_message_t message;
  int i;
  static char szText[20];

  digitalWrite(LED_BRD, LOW);
  if (twai_receive(&message, pdMS_TO_TICKS(10000)) == ESP_OK) {
    Serial.println("Message received\n");
  } else {
    Serial.println("Failed to receive CAN message\n");
    return;
  }

  //Process received message
  if (message.extd) {
    Serial.println("Message is in Extended Format");
  } else {
    Serial.println("Message is in Standard Format");
  }
  Serial.printf("vReadingCan() --> ID is 0x%X, Address: 0x%X\n\r", message.identifier, nBrdAddress);
  Serial.printf("message.identifier & 0x03FF = 0x%X\n\r", message.identifier & 0x03FF);
  if (((message.identifier & 0x03FF) == (word)nBrdAddress) || (message.identifier == CAN_BROADCAST_STOREY_BUTTON_BOARD)) {
    if(message.identifier == CAN_BROADCAST_STOREY_BUTTON_BOARD){
      Serial.println("CAN broadcast message for all storey boards.");
    }else Serial.println("CAN message is for this storey board.");
    if (!(message.rtr)) {
      Serial.println("message.rtr = 0");
      for (i = 0; i < message.data_length_code; i++) {
        Serial.printf("Data byte %d = 0x%X = %c\n\r", i, message.data[i], message.data[i]);
        if(i>0) szText[i - 1] = (char)message.data[i];
      }
      szText[i - 1] = '\0';
      // PA_PRINT
      ledMatrix.displayClear();
      //ledMatrix.displayScroll(szT, PA_CENTER, (bPuja) ? PA_SCROLL_UP : PA_SCROLL_DOWN, 100);
      //sprintf(szText,"%s",szText);
      // ledMatrix.displayScroll(szText, PA_CENTER, PA_PRINT, 100);
      switch (message.data[0]) {
        case 0: ledMatrix.setTextAlignment(PA_CENTER); // PA_LEFT, PA_RIGHT
          ledMatrix.displayText(szText, PA_CENTER,  80, 1000, PA_NO_EFFECT, PA_NO_EFFECT);
          break;
        case 1: ledMatrix.displayScroll(szText, PA_CENTER, PA_SCROLL_UP, 100);
          break;
        case 2: ledMatrix.displayScroll(szText, PA_CENTER, PA_SCROLL_DOWN, 100);
          break;
        case 3: ledMatrix.displayScroll(szText, PA_CENTER, PA_SCROLL_RIGHT, 100);
          break;
        case 4: ledMatrix.displayScroll(szText, PA_CENTER, PA_SCROLL_LEFT, 100);
          break;
        case 5: ledMatrix.displayScroll("Pendent de fer taula", PA_CENTER, PA_SCROLL_LEFT, 100);
          break;
      }

      Serial.printf("Missatge complet: \"%s\"\n\r", szText);
    } else {
      Serial.println("message.rtr = 1");
      // nCanSendAnswer();
      Serial.println("It should be --> nCanSendAnswer();");
    }
  } else if ((message.identifier & 0x03FF) == ((word)nBrdAddress | CAN_PREFIX_LED_BUTTON_COMMAND)){
    if (!(message.rtr)) {
      Serial.println("message.rtr = 0. Changing LED Buttons state");
      for (i = 0; i < message.data_length_code; i++) {
        Serial.printf("Data byte %d = 0x%X = %c\n\r", i, message.data[i], message.data[i]);
      }
      if(message.data[0] == 1) digitalWrite(EXT_LED_BT1,message.data[1]);
      if(message.data[0] == 2) digitalWrite(EXT_LED_BT2,message.data[1]);
    }
  } else {
    Serial.println("CAN message is not for this storey board.");
  }
  /*
    data[0]: REG_WR (b4: 0 (A rib). b4: 1 (B rib).)
    data[1]: Data byte if REG_WR b3 is 0. 0 or 1 if REG_WR b3 is 1.
  */
  digitalWrite(LED_BRD, HIGH);
}

int CANsend (bool bBot1,bool bBot2) {
  static unsigned int n = 250;
  static byte byOutA = 0xFF, byOutB = 0x00, byMask = 0x01;

  digitalWrite(LED_BRD, LOW);
  n++;

  byOutA = ~byMask;
  byOutB = byMask;
  byMask = (byMask == 0x80) ? 0x01 : byMask << 1;

  twai_message_t message;
  message.identifier = CAN_CARRIER_ID;
  // message.identifier = nBrdAddress; <-- Col·lisiona
  message.extd = 0;
  message.rtr = 0;
  message.data_length_code = 8;
  message.data[0] = nBrdAddress;
  message.data[1] = (bBot1)?1:0;
  message.data[2] = (bBot2)?1:0;
  message.data[3] = 0;
  message.data[4] = 0;
  message.data[5] = 0;
  message.data[6] = (byte)(n & 0xFF);
  message.data[7] = (byte)(n >> 8);
  //return twai_transmit(&message, 0);

  if (twai_transmit(&message, pdMS_TO_TICKS(1000)) == ESP_OK) {
    Serial.println("Message queued for transmission");

    if (message.extd) {
      Serial.println("Message is in Extended Format\n");
    } else {
      Serial.println("Message is in Standard Format\n");
    }
    Serial.printf("CANsend(void) --> ID is 0x%X\n", message.identifier);
    if (!(message.rtr)) {
      for (int i = 0; i < message.data_length_code; i++) {
        Serial.printf("Data byte %d = 0x%X\n", i, message.data[i]);
      }
    }

    return 1;
  } else {
    Serial.println("Failed to queue message for transmission");
    return 0;
  }
  digitalWrite(LED_BRD, HIGH);
}

void vTaskReadingCan(void *pvParameters) {
  (void) pvParameters;

  vSetupCan();

  for (;;) {
    vReadCan();
    vTaskDelay(MINIMUM_DELAY_ALLOWING_MULTITASKING);  // one tick delay (15ms) in between reads for stability
  }
}

void TaskLedW(void *pvParameters)  // This is a task.
{
  (void) pvParameters;
  bool bLedBlancEnces = LOW;

  // pinMode(LED_BRD, OUTPUT);

  for (;;) {
    bLedBlancEnces = ! bLedBlancEnces;
    // digitalWrite(LED_BRD, bLedBlancEnces);
    vTaskDelay(250);
    vTaskDelay(MINIMUM_DELAY_ALLOWING_MULTITASKING);  // one tick delay (15ms) in between reads for stability
  }
}

void TaskReadButtons(void *pvParameters)  // This is a task.
{
  (void) pvParameters;
  String str,oldStr = "";
  bool bBt1,bBt2;
  
  for (;;) {
    bBt1 = !digitalRead(BT1);
    bBt2 = !digitalRead(BT2);
    str  = (bBt1)?"BT1 premut":"BT1 no premut";
    str += (bBt2)?", BT2 premut":", BT2 no premut";
    if(str != oldStr){
      oldStr = str;
      if(bBt1 || bBt2) CANsend(bBt1,bBt2);
      Serial.println(str);
    } 
    // digitalWrite(EXT_LED_BT1,!digitalRead(BT1));
    // digitalWrite(EXT_LED_BT2,!digitalRead(BT2));
    if(!digitalRead(BT1)) digitalWrite(EXT_LED_BT1,HIGH);
    if(!digitalRead(BT2)) digitalWrite(EXT_LED_BT2,HIGH);
    vTaskDelay(MINIMUM_DELAY_ALLOWING_MULTITASKING);  // one tick delay (15ms) in between reads for stability
  }

}
void setup() {
  Serial.begin(115200);
  Serial.println(__FILE__);

  vSetupIO();
  vSetupLedMatrix();
  nBrdAddress = byReadAddress();
  Serial.print("Address: 0x");
  Serial.println(nBrdAddress, HEX);

  ledMatrix.displayScroll("Ascensors Jorda", PA_CENTER, PA_SCROLL_LEFT, 100);

  xTaskCreatePinnedToCore(
    TaskReadButtons
    ,  "TaskReadButtons"   // A name just for humans
    ,  2 * 1024 // This stack size can be checked & adjusted by reading the Stack Highwater
    ,  NULL
    ,  1  // Priority, with 3 (configMAX_PRIORITIES - 1) being the highest, and 0 being the lowest.
    ,  NULL
    ,  ARDUINO_RUNNING_CORE);

  xTaskCreatePinnedToCore(
    TaskLedW
    ,  "TaskLedW"   // A name just for humans
    ,  2 * 1024 // This stack size can be checked & adjusted by reading the Stack Highwater
    ,  NULL
    ,  1  // Priority, with 3 (configMAX_PRIORITIES - 1) being the highest, and 0 being the lowest.
    ,  NULL
    ,  ARDUINO_RUNNING_CORE);
  //
  xTaskCreatePinnedToCore(
    vTaskReadingCan
    ,  "vTaskReadingCan"   // A name just for humans
    ,  8 * 1024 // This stack size can be checked & adjusted by reading the Stack Highwater
    ,  NULL
    ,  2  // Priority, with 3 (configMAX_PRIORITIES - 1) being the highest, and 0 being the lowest.
    ,  NULL
    ,  ARDUINO_RUNNING_CORE);

}

void loop() {
  long int liT = millis();
  static bool sbPrimerCop = true, bPuja = true;
  static int sbyN = 0;
  static char szT[12];
  static long int sliTp = liT;

  if (liT < 8000) {
    // Display "Ascensors Jorda"
  } else {
    if (sbPrimerCop) {
      sbPrimerCop = false;
      //ledMatrix.begin();
      //ledMatrix.setTextAlignment(PA_CENTER); // PA_LEFT, PA_RIGHT
    } else {
      /*
        //ledMatrix.print(sbyN++);
        if (liT / 2000 != sliTp / 2000) {
          sliTp = liT;
          if (sbyN) sprintf(szT, "%ld", sbyN);
          else sprintf(szT, "B", sbyN);
          // Serial.print(szT);
          ledMatrix.displayClear();
          ledMatrix.displayScroll(szT, PA_CENTER, (bPuja) ? PA_SCROLL_UP : PA_SCROLL_DOWN, 100);
          sbyN = (bPuja) ? sbyN + 1 : sbyN - 1;
          if (sbyN > 4) {
            bPuja = false;
            sbyN--;
          }
          if (sbyN < -2) {
            bPuja = true;
            sbyN++;
          }
        }
        //delay(250);
      */
    }
  }

  if (ledMatrix.displayAnimate()) {
    ledMatrix.displayReset();
  }
  //vReadingCan();
}
